from __future__ import annotations

import functools
import typing

from metadsl import *
from metadsl_core import *

from .boxing import *

__all__ = ["injest", "guess_type", "guess_types", "guess_type_of_type"]


T = typing.TypeVar("T")
U = typing.TypeVar("U")


def injest(o: object) -> object:
    """
    Returns a metadsl expression that is roughly compatible with the Python object.

    Shouldn't be used in real code.
    """
    compat_tp, inner_tp = guess_type(o)
    return Boxer[compat_tp, inner_tp].convert_and_box(o)  # type: ignore


@functools.singledispatch
def guess_type(o: object) -> typing.Tuple[typing.Type, typing.Type]:
    """
    Takes in an object and returns a typle of the outer compat type and inner type.
    """
    raise NotImplementedError(
        f"Don't know how to guess type {type(o)}, provide an explicit type to injest."
    )


def guess_types(o: object, *os: object) -> typing.Tuple[typing.Type, typing.Type]:
    """
    Guesses all and makes sure they have they have the same types
    """
    compat_tp, inner_tp = guess_type(o)
    for o in os:
        new_compat_tp, new_inner_tp = guess_type(o)
        if (compat_tp != new_compat_tp) or (inner_tp != new_inner_tp):
            raise NotImplementedError(f"different guesses for {(o,) + os}")
    return compat_tp, inner_tp


class CreateInstance(Expression, typing.Generic[T]):
    @expression
    @classmethod
    def create(cls) -> T:
        ...


def create_instance(tp: typing.Type[T]) -> T:
    return CreateInstance[tp].create()  # type: ignore


def guess_type_of_type(tp: typing.Type) -> typing.Tuple[typing.Type, typing.Type]:
    """
    Guesses the type of a type, by creating a dummy instance of it and guessing the type of that.

    We need this b/c functools.singledispatch doesn't really work at the type level, so we just make
    dummy instances instead.
    """
    return guess_type(create_instance(tp))



# from __future__ import annotations

# from metadsl import *
# from metadsl_core import *
# import typing

# IntType = typing.Union[int, Integer]
# IntsType = typing.Union[typing.Tuple[int, ...], Vec[Integer]]
# IdxType = typing.Union[IntType, Vec[Integer], typing.Tuple[int, ...]]
# ArrayType = typing.Union[intType, ]

# class ArrayCompat(Expression):


#     def __getitem__(self, idx: IdxType) -> ArrayCompat:
#         ...

#     def __setitem__(self, idx: IdxType, )


from __future__ import annotations
from metadsl import *
from metadsl_core import *
from metadsl_rewrite import *

from .bool_compat import *
from .injest import *


t = BoolCompat.from_maybe_boolean(Maybe.just(Boolean.create(True)))
f = BoolCompat.from_maybe_boolean(Maybe.just(Boolean.create(False)))


def test_injest() -> None:
    assert execute(injest(True)) == t
    assert execute(injest(Boolean.create(True))) == t
    assert execute(injest(t)) == t


def test_if_bool() -> None:
    assert execute(t.if_(True, False)) == t
    assert execute(f.if_(True, False)) == f



from __future__ import annotations

import metadsl.typing_tools
import pytest
from metadsl import *

from metadsl_rewrite import *

from .abstraction import *
from .conversion import *
from .function import *
from .integer import *
from .maybe import *
from .strategies import *

one = Integer.from_int(1)
zero = Integer.from_int(0)


class TestFunction:
    def test_zero_abstraction(self):
        @FunctionZero.from_fn
        def fn() -> Integer:
            return one

        assert execute(fn()) == one

    def test_one_abstraction(self):
        @FunctionOne.from_fn
        def add_one(a: Integer) -> Integer:
            return a + one

        assert (
            metadsl.typing_tools.get_type(execute(add_one.abstraction))
            == Abstraction[Integer, Integer]
        )

        assert execute(add_one.abstraction(zero)) == one

    def test_one_recursive_call(self):
        @FunctionOne.from_fn_recursive
        def factorial(fact_fn: FunctionOne[Integer, Integer], n: Integer) -> Integer:
            return n.eq(zero).if_(one, n * fact_fn(n - one))

        assert execute(factorial(zero)) == one
        assert execute(factorial(one)) == one
        assert execute(factorial(Integer.from_int(3))) == Integer.from_int(6)

    def test_two_abstraction(self):
        @FunctionTwo.from_fn
        def add(a: Integer, b: Integer) -> Integer:
            return a + b

        assert (
            metadsl.typing_tools.get_type(execute(add.abstraction))
            == Abstraction[Integer, Abstraction[Integer, Integer]]
        )

        assert execute(add.abstraction(one)(zero)) == one

    def test_two_call(self):
        @FunctionTwo.from_fn
        def add(a: Integer, b: Integer) -> Integer:
            return a + b

        assert execute(
            add(Integer.from_int(1), Integer.from_int(2))
        ) == Integer.from_int(3)

    def test_three_recursive_call(self):
        @FunctionThree.from_fn_recursive
        def fib_more(
            fn: FunctionThree[Integer, Integer, Integer, Integer],
            n: Integer,
            a: Integer,
            b: Integer,
        ) -> Integer:
            pred_cont = n > one
            minus1 = n - one
            ab = a + b
            added = fn(minus1, b, ab)

            n_eq_1 = n.eq(one)
            return pred_cont.if_(added, n_eq_1.if_(b, a))

        @FunctionOne.from_fn
        def fib(n: Integer) -> Integer:
            return fib_more(n, zero, one)

        assert execute(fib(zero)) == zero
        assert execute(fib(one)) == one
        assert execute(fib(Integer.from_int(8))) == Integer.from_int(21)


class T(Expression):
    ...


class U(Expression):
    ...


class V(Expression):
    ...


class X(Expression):
    ...


@expression
def t_to_u(t: T) -> U:
    ...


@expression
def t() -> T:
    ...


@expression
def v() -> V:
    ...


@expression
def v_to_t(v: V) -> T:
    ...


@expression
def u_to_x(u: U) -> X:
    ...


@rule
def convert_t_to_u_just(t: T) -> R[Maybe[U]]:
    return Converter[U].convert(t), Maybe.just(t_to_u(t))


@rule
def convert_t_to_u_nothing(t: T) -> R[Maybe[U]]:
    return Converter[U].convert(t), Maybe[U].nothing()


def fn() -> T:
    return t()


def fn_one(t: T) -> U:
    return t_to_u(t)


@rule
def convert_v_to_t_just(v: V) -> R[Maybe[T]]:
    return Converter[T].convert(v), Maybe.just(v_to_t(v))


@rule
def convert_v_to_t_nothing(v: V) -> R[Maybe[T]]:
    return Converter[T].convert(v), Maybe[T].nothing()


@rule
def convert_u_to_x_just(u: U) -> R[Maybe[X]]:
    return Converter[X].convert(u), Maybe.just(u_to_x(u))


@rule
def convert_u_to_x_nothing(u: U) -> R[Maybe[X]]:
    return Converter[X].convert(u), Maybe[X].nothing()


class TestFunctionConversion:
    @pytest.mark.parametrize(
        "t_to_u_rule",
        [convert_t_to_u_just, convert_t_to_u_nothing],
        ids=["just", "nothing"],
    )
    @pytest.mark.parametrize(
        "input",
        [fn, FunctionZero[T].from_fn(fn)],
        ids=["python function", "FunctionZero"],
    )
    def test_zero(self, input, t_to_u_rule):
        result = Converter[FunctionZero[Maybe[U]]].convert(input)
        desired = Maybe.just(FunctionZero.create("fn", Converter[U].convert(t())))
        with register.tmp(t_to_u_rule):
            assert execute(result) == execute(desired)

    @pytest.mark.parametrize(
        "u_to_x_rule",
        [convert_v_to_t_just, convert_u_to_x_nothing],
        ids=["just", "nothing"],
    )
    @pytest.mark.parametrize(
        "v_to_t_rule",
        [convert_v_to_t_just, convert_v_to_t_nothing],
        ids=["just", "nothing"],
    )
    @pytest.mark.parametrize(
        "input",
        [fn_one, FunctionOne[T, U].from_fn(fn_one)],
        ids=["python function", "FunctionOne"],
    )
    def test_one(self, input, u_to_x_rule, v_to_t_rule):
        converted_fn = Converter[FunctionOne[V, Maybe[X]]].convert(input)
        result: Maybe[X] = converted_fn.flat_map(
            Abstraction[FunctionOne[V, Maybe[X]], Maybe[X]].from_fn(lambda fn: fn(v()))
        )

        desired: Maybe[X] = Converter[T].convert(v()).map(
            Abstraction[T, U].from_fn(fn_one)
        ).flat_map(Abstraction[U, Maybe[X]].from_fn(Converter[X].convert))

        with register.tmp(u_to_x_rule, v_to_t_rule):
            assert execute(result) == execute(desired)



from __future__ import annotations
import typing
import types
import functools
import typing_inspect

from metadsl import *
from metadsl_core import *
from metadsl.typing_tools import ToCallable, get_type
from metadsl_rewrite import *

from .injest import *
from .boxing import *

__all__: typing.List[str] = []

T = typing.TypeVar("T")
U = typing.TypeVar("U")

V = typing.TypeVar("V")
X = typing.TypeVar("X")


# class FunctionZeroCompat(Expression, typing.Generic[T, U]):
#     @expression
#     @classmethod
#     def from_maybe_function(
#         cls, fn: Maybe[FunctionZero[Maybe[U]]]
#     ) -> FunctionOneCompat[T, U]:
#         ...

#     @expression
#     @classmethod
#     def from_function(cls, fn: FunctionZero[T]) -> FunctionOneCompat[T, U]:
#         ...

#     @expression
#     def __call__(self, arg: object) -> T:
#         ...


class FunctionOneCompat(Expression, typing.Generic[T]):
    """
    Compat function with inner function mapping from U -> Maybe[V]

    that converts any arg to U and then boxes the result into the compat type T.
    """

    @expression
    @classmethod
    def box(cls, fn: Maybe[FunctionOne[U, Maybe[V]]]) -> FunctionOne[object, T]:
        ...


@register_convert
@rule
def box_function_one(v: Maybe[FunctionOne[U, Maybe[V]]]) -> R[FunctionOne[object, T]]:
    return (
        Boxer[FunctionOne[object, T], FunctionOne[U, Maybe[V]]].box(v),
        FunctionOneCompat[T].box(v),
    )


@expression
def converter_name(name: str) -> str:
    return name + "_convert"


register_core(default_rule(converter_name))


@register_convert
@rule
def function_one_box_rule(fn: FunctionOne[U, Maybe[V]]) -> R[FunctionOne[object, T]]:

    return (
        FunctionOneCompat[T].box(Maybe.just(fn)),
        lambda: FunctionOne[object, T].from_fn(
            lambda o: Boxer[T, V].box(Converter[U].convert(o).flat_map(fn.abstraction)),
            name=Maybe.just(converter_name(fn.name)),
        ),
    )


# Register for anything that could be thought of as a callable, but don't use typing.Callable
# since then callable-ish things would be supported like  FunctionOne themselves
@guess_type.register(ToCallable)
def guess_python_function_tp(
    fn: typing.Callable,
) -> typing.Union[
    typing.Tuple[typing.Type[FunctionZero], typing.Type[FunctionZero]],
    typing.Tuple[typing.Type[FunctionOne], typing.Type[FunctionOne]],
]:
    """
    Guesses the type of a function. The args of the function should
    all be compat types.
    """
    arg_tps, return_tp = typing_inspect.get_args(get_type(fn))

    return_compat_tp, return_inner_tp = guess_type_of_type(return_tp)

    arg_inner_tps = [guess_type_of_type(arg_tp)[1] for arg_tp in arg_tps]
    n_args = len(arg_inner_tps)

    if n_args == 0:
        raise NotImplementedError
        # return (
        #     FunctionZeroCompat[return_compat_tp, return_inner_tp],
        #     FunctionZero[Maybe[return_inner_tp]],
        # )
    if n_args == 1:
        return (
            FunctionOne[object, return_compat_tp],  # type: ignore
            FunctionOne[arg_inner_tps[0], Maybe[return_inner_tp]],  # type: ignore
        )
    raise NotImplementedError


@guess_type.register
def guess_function_one_tp(fn: FunctionOne):
    arg_tp, ret_tp = typing_inspect.get_args(get_type(fn))
    ret_compat_tp, ret_inner_tp = guess_type_of_type(ret_tp)

    return FunctionOne[object, ret_compat_tp], FunctionOne[arg_tp, Maybe[ret_inner_tp]]  # type: ignore


# TODO:
# * Add box logic for maybe x, y to FunctionOne



"""
Library to help create DSLs in Python.
"""

__version__ = "0.4.0"

from .dict_tools import *
from .expressions import *
from .logging import *
from .module_tools import *
from .normalized import *

export_from(
    __name__, "expressions", "dict_tools", "normalized", "logging", "module_tools"
)



"""
Useful for to convert object to boxed types.
"""
from __future__ import annotations

import typing

from metadsl import *
from metadsl_rewrite import *

from .abstraction import *
from .maybe import *
from .pair import *
from .strategies import *

__all__ = ["Converter", "convert_identity_rule", "convert_to_maybe"]

T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")
X = typing.TypeVar("X")


class Converter(Expression, typing.Generic[T]):
    @expression
    @classmethod
    def convert(cls, value: object) -> Maybe[T]:
        """
        Converts a value to a certain type.
        """
        ...


@register_convert
@rule
def convert_identity_rule(value: T) -> R[Maybe[T]]:
    """
    When the value is an instance of the type being converted, we can convert it.
    """
    yield Converter[T].convert(value), Maybe.just(value)
    yield Converter[T].convert(Maybe.just(value)), Maybe.just(value)
    yield Converter[T].convert(Maybe[T].nothing()), Maybe[T].nothing()


@register_convert
@rule
def convert_to_maybe(x: object) -> R[Maybe[Maybe[T]]]:
    return (
        Converter[Maybe[T]].convert(x),
        lambda: Maybe.just(
            Maybe[T].nothing() if x is None else Converter[T].convert(x)
        ),
    )


@register_convert
@rule
def convert_to_pair(l: object, r: object) -> R[Maybe[Pair[T, U]]]:
    return (
        Converter[Pair[T, U]].convert(Pair.create(l, r)),
        Converter[T].convert(l) & Converter[U].convert(r),
    )


@expression
def lift_abstraction_maybe(a: Abstraction[T, Maybe[U]]) -> Maybe[Abstraction[T, U]]:
    """
    Deprecated: This is a bad idea.

    This would require some kinda of global match context. i.e. imagine matchign on the result,
    in the just case, then you have a fn you call on two args... One ends up resulting in true, the other false,
    then you need to know that whole match body was false, and to actually match on the nothing result.
    """
    ...


@register_core
@rule
def lift_abstraction_maybe_rule(v: T, b: U) -> R[Maybe[Abstraction[T, U]]]:
    # Need to hard code in cases, b/c we don't wanna lift unbound variable outside of abstraction
    yield (
        lift_abstraction_maybe(Abstraction.create(v, Maybe.just(b))),
        Maybe.just(Abstraction.create(v, b)),
    )
    yield (
        lift_abstraction_maybe(Abstraction.create(v, Maybe[U].nothing())),
        Maybe[Abstraction[T, U]].nothing(),
    )


@register_convert
@rule
def convert_to_abstraction(a: Abstraction[T, U]) -> R[Maybe[Abstraction[V, Maybe[X]]]]:
    """
    Converting from one abstraction to another means putting a convert on the input and the output.
    """

    return (
        Converter[Abstraction[V, Maybe[X]]].convert(a),
        lambda: Maybe.just(
            Abstraction[V, Maybe[X]].from_fn(
                lambda v: Converter[T]
                .convert(v)
                .flat_map(
                    Abstraction[T, Maybe[X]].from_fn(
                        lambda t: Converter[X].convert(a(t))
                    )
                )
            )
        ),
    )


@register_convert
@rule
def convert_fn_to_abstraction(
    a: typing.Callable[[T], U]
) -> R[Maybe[Abstraction[V, Maybe[X]]]]:
    return (
        Converter[Abstraction[V, Maybe[X]]].convert(a),
        lambda: Converter[Abstraction[V, Maybe[X]]].convert(Abstraction.from_fn(a)),
    )



from __future__ import annotations

import typing

from metadsl import *
from metadsl_rewrite import *

from .abstraction import *
from .boolean import *
from .conversion import *
from .maybe import *
from .strategies import *

__all__ = ["Integer"]

T = typing.TypeVar("T")


class Integer(Expression):
    @expression
    @classmethod
    def from_int(cls, i: int) -> Integer:
        ...

    @expression
    def __add__(self, other: Integer) -> Integer:
        ...

    @expression
    def __sub__(self, other: Integer) -> Integer:
        ...

    @expression
    def __mul__(self, other: Integer) -> Integer:
        ...

    @expression
    def __floordiv__(self, other: Integer) -> Integer:
        ...

    @expression
    def __mod__(self, other: Integer) -> Integer:
        ...

    @expression
    def eq(self, other: Integer) -> Boolean:
        ...

    @expression
    def __lt__(self, other: Integer) -> Boolean:
        ...

    @expression
    def __le__(self, other: Integer) -> Boolean:
        ...

    @expression
    def __gt__(self, other: Integer) -> Boolean:
        ...

    @expression
    def __ge__(self, other: Integer) -> Boolean:
        ...

    @expression
    def fold(self, initial: T, fn: Abstraction[Integer, Abstraction[T, T]]) -> T:
        ...

    @expression
    @classmethod
    def zero(cls) -> Integer:
        return Integer.from_int(0)

    @expression
    @classmethod
    def one(cls) -> Integer:
        return cls.zero().inc

    @expression  # type: ignore
    @property
    def inc(self) -> Integer:
        return self + Integer.from_int(1)

    @expression
    @classmethod
    def infinity(cls) -> Integer:
        ...


register_ds(default_rule(Integer.zero))
register_ds(default_rule(Integer.inc))
register_ds(default_rule(Integer.one))


@register_ds
@rule
def integer_math(l: int, r: int) -> R[Integer]:
    yield Integer.from_int(l) + Integer.from_int(r), lambda: Integer.from_int(l + r)
    yield Integer.from_int(l) - Integer.from_int(r), lambda: Integer.from_int(l - r)
    yield Integer.from_int(l) * Integer.from_int(r), lambda: Integer.from_int(l * r)
    yield Integer.from_int(l) // Integer.from_int(r), lambda: Integer.from_int(l // r)
    yield Integer.from_int(l) % Integer.from_int(r), lambda: Integer.from_int(l % r)

    inf = Integer.infinity()
    yield Integer.from_int(l) + inf, inf
    yield inf + Integer.from_int(r), inf


@register_ds
@rule
def integer_comparison(l: int, r: int) -> R[Boolean]:
    li = Integer.from_int(l)
    ri = Integer.from_int(r)
    yield li.eq(ri), lambda: Boolean.create(l == r)
    yield li < ri, lambda: Boolean.create(l < r)
    yield li <= ri, lambda: Boolean.create(l <= r)
    yield li > ri, lambda: Boolean.create(l > r)
    yield li >= ri, lambda: Boolean.create(l >= r)

    inf = Integer.infinity()
    true = Boolean.true()
    false = Boolean.false()

    yield li.eq(inf), false
    yield inf.eq(ri), false
    yield li < inf, true
    yield inf < ri, false
    yield li <= inf, true
    yield inf <= ri, false
    yield li > inf, false
    yield inf > ri, true
    yield li >= inf, false
    yield inf >= ri, true


@register_ds
@rule
def fold(i: int, initial: T, fn: Abstraction[Integer, Abstraction[T, T]]) -> R[T]:
    def inner() -> T:
        value = initial
        for i_ in range(i):
            value = fn(Integer.from_int(i_))(value)
        return value

    return Integer.from_int(i).fold(initial, fn), inner


@register_convert
@rule
def convert_integer(i: int) -> R[Maybe[Integer]]:
    return Converter[Integer].convert(i), Maybe.just(Integer.from_int(i))



from __future__ import annotations

import typing

from metadsl import *
from metadsl_rewrite import *

from .boolean import *
from .integer import *
from .maybe import *
from .conversion import *
from .abstraction import *
from .pair import *
from .strategies import *

T = typing.TypeVar("T")
U = typing.TypeVar("U")

__all__ = ["Vec", "Selection"]


class NoValue(Expression, typing.Generic[T]):
    """
    Returned by a getitem function that can have no inputs
    """

    @expression
    @classmethod
    def create(cls) -> T:
        ...


class Selection(Expression):
    """
    A selection represents a mapping from a vector to a new vector, that includes some subset of the old
    values, but in new positions. Old values can be duplicated.
    """

    @expression
    @classmethod
    def create_slice(cls, start: Integer, stop: Integer, step: Integer) -> Selection:
        """
        Like a python slice, but no negatives. start <= i < stop where i % step == start
        """
        ...

    @expression
    @classmethod
    def create_slice_optional(
        cls, start: Maybe[Integer], stop: Maybe[Integer], step: Maybe[Integer]
    ) -> Selection:
        return cls.create_slice(
            start.default(Integer.zero()),
            stop.default(Integer.infinity()),
            step.default(Integer.one()),
        )

    @expression
    @classmethod
    def create_indices(cls, indices: Vec[Integer]) -> Selection:
        """
        Selection indices from this vector.
        """
        ...

    @expression
    def old_to_new(self, idx: Integer) -> Vec[Integer]:
        """
        Map from old index to where it shows up in new vector.
        """
        ...

    @expression
    def new_to_old(self, idx: Integer) -> Integer:
        """
        Map from new index to the old index it came from.
        """
        ...

    @expression
    def length(self, old_length: Integer) -> Integer:
        """
        Length of new parts.
        """
        ...


register_ds(default_rule(Selection.create_slice_optional))


class Vec(Expression, typing.Generic[T]):
    """
    A tuple of homogonous types.
    """

    @expression
    def __getitem__(self, index: Integer) -> T:
        ...

    @expression
    def select(self, selection: Selection) -> Vec[T]:
        ...

    # @expression
    # def full_index

    # @expression
    # def full_single_index

    @expression
    @classmethod
    def create(cls, *items: T) -> Vec[T]:
        ...

    @expression
    def map(self, fn: Abstraction[T, U]) -> Vec[U]:
        ...

    @expression
    def append(self, x: T) -> Vec[T]:
        ...

    @expression
    def fold(self, initial: U, fn: Abstraction[U, Abstraction[T, U]]) -> U:
        """
        Like this:

            def fold(xs, initial, fn):
                ret = initial
                for x in xs:
                    ret = fn(ret)(x)
                return ret
        """

    @expression
    def __add__(self, other: Vec[T]) -> Vec[T]:
        ...

    @expression  # type: ignore
    @property
    def length(self) -> Integer:
        ...

    @expression
    def take(self, i: Integer) -> Vec[T]:
        ...

    @expression
    def drop(self, i: Integer) -> Vec[T]:
        ...

    @expression
    @classmethod
    def create_fn(cls, length: Integer, fn: Abstraction[Integer, T]) -> Vec[T]:
        ...

    @expression
    @classmethod
    def lift_maybe(cls, vec: Vec[Maybe[T]]) -> Maybe[Vec[T]]:
        ...

    @expression
    def set(self, i: Integer, value: T) -> Vec[T]:
        ...

    @expression
    def set_selection(self, s: Selection, values: Vec[T]) -> Vec[T]:
        ...

    @expression  # type: ignore
    @property
    def empty(self) -> Boolean:
        return self.length.eq(Integer.zero())

    @expression  # type: ignore
    @property
    def first(self) -> Maybe[T]:
        return self.empty.if_(Maybe[T].nothing(), Maybe.just(self[Integer.zero()]))


register_ds(default_rule(Vec[T].empty))
register_ds(default_rule(Vec[T].first))


@register_ds
@rule
def create_fn_rules(
    l: Integer,
    fn: Abstraction[Integer, T],
    x: T,
    l1: Integer,
    i: Integer,
    fn1: Abstraction[Integer, T],
    n: Integer,
    s: Selection,
):
    v = Vec.create_fn(l, fn)
    v1 = Vec.create_fn(l1, fn1)
    one = Integer.from_int(1)
    yield v[i], fn(i)
    yield v.append(x), Vec.create_fn(
        l + one, Abstraction[Integer, T].from_fn(lambda i: (i < l).if_(fn(i), x))
    )
    yield v + v1, Vec.create_fn(
        l + l1,
        Abstraction[Integer, T].from_fn(lambda i: (i < l).if_(fn(i), fn1(i - l1))),
    )
    yield v.length, l
    yield v.take(n), Vec.create_fn(l - n, fn)
    yield v.drop(n), Vec.create_fn(
        l - n, Abstraction[Integer, T].from_fn(lambda i: fn(i + n)),
    )
    yield v.select(s), Vec.create_fn(
        s.length(l), Abstraction[Integer, T].from_fn(lambda i: v[s.new_to_old(i)]),
    )
    yield v.set(n, x), Vec.create_fn(
        l, Abstraction[Integer, T].from_fn(lambda i: i.eq(n).if_(x, fn(i)))
    )
    # set selection maps each old index, looking up the first new one on the new v, if there
    # are any, otherwise, use old v
    yield v.set_selection(s, v1), Vec.create_fn(
        l,
        Abstraction[Integer, T].from_fn(
            lambda i: s.old_to_new(i).first.match(fn(i), fn1)
        ),
    )


@register_ds
@rule
def select_slice(start: Integer, stop: Integer, step: Integer, i: Integer):
    s = Selection.create_slice(start, stop, step)
    yield (
        s.old_to_new(i),
        # If the index is less than the stop index, and is on the step from the start,
        # return the new one, otherwise return none
        (i < stop)
        .and_((i % step).eq(start))
        .if_(Vec.create((i - start) // step), Vec[Integer].create()),
    )

    yield (s.new_to_old(i), start + (i * step))
    # Compute the actual stop if one is specified, by taking min of stop and old length "i"
    actual_stop = (i < stop).if_(i, stop)
    yield (s.length(i), (stop - actual_stop) // step)


@register_ds  # type: ignore
@rule
def length_rule(xs: typing.Sequence[T]) -> R[Integer]:
    return (Vec[T].create(*xs).length, lambda: Integer.from_int(len(xs)))


@register_ds  # type: ignore
@rule
def take_drop_rule(xs: typing.Sequence[T], i: int) -> R[Vec[T]]:
    yield (
        Vec[T].create(*xs).take(Integer.from_int(i)),
        lambda: Vec[T].create(*xs[:i]),
    )
    yield (
        Vec[T].create(*xs).drop(Integer.from_int(i)),
        lambda: Vec[T].create(*xs[i:]),
    )


@register_ds  # type: ignore
@rule
def add_rule(ls: typing.Sequence[T], rs: typing.Sequence[T]) -> R[Vec[T]]:
    return (Vec[T].create(*ls) + Vec[T].create(*rs), Vec[T].create(*ls, *rs))


@register_ds  # type: ignore
@rule
def getitem(i: int, xs: typing.Sequence[T]) -> R[T]:
    return (Vec[T].create(*xs)[Integer.from_int(i)], lambda: xs[i])


@register_ds
@rule
def map(fn: Abstraction[T, U], xs: typing.Sequence[T]) -> R[Vec[U]]:
    return (
        Vec[T].create(*xs).map(fn),  # type: ignore
        lambda: (Vec.create(*(fn(x) for x in xs))) if xs else Vec[U].create(),
    )


@register_ds
@rule
def append(xs: typing.Sequence[T], x: T) -> R[Vec[T]]:
    return (Vec[T].create(*xs).append(x), lambda: Vec.create(*xs, x))


# TODO: Replace with exact replacement for slice
@register_ds
@rule
def vec_select(xs: typing.Sequence[T], s: Selection) -> R[Vec[T]]:
    v = Vec[T].create(*xs)

    def inner() -> Vec[T]:
        # contains a list of of indices that we should lookup in the current vec
        result_idxs = Vec[Integer].create()
        for i in range(len(xs)):
            result_idxs += s.old_to_new(Integer.from_int(i))
        # naw just do a fold here, oh well!

        # TODO: Should this just be a Vector.from_fn instead?
        return result_idxs.fold(
            Vec[T].create(),
            Abstraction[Vec[T], Abstraction[Integer, Vec[T]]].from_fn(
                lambda result: Abstraction[Integer, Vec[T]].from_fn(
                    lambda i: result.append(v[i])
                )
            ),
        )

    return v.select(s), inner


@register_ds
@rule
def vec_set(xs: typing.Sequence[T], i: int, value: T) -> R[Vec[T]]:
    return (
        Vec[T].create(*xs).set(Integer.from_int(i), value),
        lambda: Vec[T].create(*(value if i == i_ else v for i_, v in enumerate(xs))),
    )


@register_ds
@rule
def vec_set_selection(
    xs: typing.Sequence[T], start: int, stop: int, step: int, values: typing.Sequence[T]
) -> R[Vec[T]]:
    def inner() -> Vec[T]:
        l = list(xs)
        l[start:stop:step] = list(values)
        return Vec[T].create(*l)

    yield (
        Vec[T]
        .create(*xs)
        .set_selection(
            Selection.create_slice(
                Integer.from_int(start), Integer.from_int(stop), Integer.from_int(step)
            ),
            Vec[T].create(*values),
        ),
        inner,
    )

    def inner_inf() -> Vec[T]:
        l = list(xs)
        l[start::step] = list(values)
        return Vec[T].create(*l)

    yield (
        Vec[T]
        .create(*xs)
        .set_selection(
            Selection.create_slice(
                Integer.from_int(start), Integer.infinity(), Integer.from_int(step)
            ),
            Vec[T].create(*values),
        ),
        inner,
    )


# def tuple_all(t: Vec[Maybe[T]]) -> Maybe[Vec[T]]:
#     # accumulate a maybe tuple of T, representng the partial list so far
#     @Abstraction.from_fn
#     def fn(tpl: Maybe[Vec[T]]) -> Abstraction[Maybe[T], Maybe[Vec[T]]]:
#         @Abstraction.from_fn
#         def inner(x: Maybe[T]) -> Maybe[Vec[T]]:
#             @Abstraction.from_fn
#             def tpl_matched(just_tpl: Vec[T]) -> Maybe[Vec[T]]:

#                 @Abstraction.from_fn
#                 def x_matched(just_x: T) -> Maybe[Vec[T]]:
#                     return just_tpl.append(just_x)

#                 return x.match(
#                     Maybe[Vec[T]].nothing(),
#                     x_matched
#                 )
#             return tpl.match(
#                 tpl,
#                 tpl_matched
#             )
#         return inner

#     return t.fold(Maybe.just(Vec[T].create()), fn)


@register_ds
@rule
def fold_vec(
    xs: typing.Sequence[T], initial: U, fn: Abstraction[U, Abstraction[T, U]]
) -> R[U]:
    def inner() -> U:
        res = initial
        for x in xs:
            res = fn(res)(x)
        return res

    return (Vec.create(*xs).fold(initial, fn), inner)


@register_convert
@rule
def convert_vec(xs: typing.Sequence[object]) -> R[Maybe[Vec[T]]]:
    def inner() -> Maybe[Vec[T]]:
        if not isinstance(xs, tuple):
            raise NoMatch
        return Vec[T].lift_maybe(
            Vec[Maybe[T]].create(*(Converter[T].convert(x) for x in xs))
        )

    return (
        Converter[Vec[T]].convert(xs),
        inner,
    )


@register_convert
@rule
def convert_slice(s: slice) -> R[Maybe[Selection]]:
    return (
        Converter[Selection].convert(s),
        lambda: (
            Converter[Maybe[Integer]].convert(s.start)
            & (
                Converter[Maybe[Integer]].convert(s.stop)
                & Converter[Maybe[Integer]].convert(s.step)
            )
        ).map(
            Abstraction[
                Pair[Maybe[Integer], Pair[Maybe[Integer], Maybe[Integer]]], Selection
            ].from_fn(
                lambda s_: Selection.create_slice_optional(
                    s_.left, s_.right.left, s_.right.right
                )
            )
        ),
    )


@register_core
@rule
def lift_maybe(v: Vec[Maybe[T]]) -> R[Maybe[Vec[T]]]:

    return (
        Vec.lift_maybe(v),
        v.fold(
            Maybe.just(Vec[T].create()),
            Abstraction[Maybe[Vec[T]], Abstraction[Maybe[T], Maybe[Vec[T]]]].from_fn(
                lambda xs: Abstraction[Maybe[T], Maybe[Vec[T]]].from_fn(
                    lambda x: (xs & x).map(
                        Abstraction[Pair[Vec[T], T], Vec[T]].from_fn(
                            lambda both: both.left.append(both.right)
                        )
                    )
                )
            ),
        ),
    )



from __future__ import annotations
import typing

from metadsl import *
from metadsl_core import *
from metadsl_rewrite import *

__all__ = ["Boxer"]

T = typing.TypeVar("T")
U = typing.TypeVar("U")


class Boxer(Expression, typing.Generic[T, U]):
    """
    Box a value into the compat type T from type U.
    """

    @expression
    @classmethod
    def box(cls, value: Maybe[U]) -> T:
        ...

    @expression
    @classmethod
    def convert(cls, value: object) -> Maybe[U]:
        return Converter[U].convert(value)

    @expression
    @classmethod
    def convert_and_box(cls, value: object) -> T:
        return cls.box(cls.convert(value))


register_convert(default_rule(Boxer[T, U].convert))
register_convert(default_rule(Boxer[T, U].convert_and_box))



# from __future__ import annotations

# import numpy
# from metadsl import *


# # TODO: Move to seperate numpy repo
# # inherit from ndarray like
# # but use "compat" types for shape etc.
# # Each should have a `to_xxx` method.
# from .vec import *
# from .integer import *
# from .conversion import *
# from .strategies import *
# from .maybe import *
# from .abstraction import *
# from .pair import *
# from .either import *


# __all__ = ["arange", "IndxType", "NDArray"]


# IndxType = Either[Integer, Vec[Integer]]


# class NDArray(Expression):
#     @expression
#     def __getitem__(self, idxs: IndxType) -> NDArray:
#         ...

#     @expression
#     def __add__(self, other: NDArray) -> NDArray:
#         ...

#     @expression
#     def to_ndarray(self) -> numpy.ndarray:
#         ...

#     @expression
#     @classmethod
#     def from_ndarray(cls, n: numpy.ndarray) -> NDArray:
#         ...

#     @expression
#     @classmethod
#     def from_ndarray(cls, n: numpy.ndarray) -> NDArray:
#         ...

# @register_unbox
# @rule
# def box_unbox_ndarray(n: numpy.ndarray) -> R[numpy.ndarray]:
#     return NDArray.from_ndarray(n).to_ndarray(), n


# class NDArrayCompat(Expression):
#     @expression
#     def __getitem__(self, idxs: object) -> NDArrayCompat:
#         return NDArrayCompat.from_ndarray(
#             (
#                 Converter[NDArray].convert(self)
#                 & Converter[Either[Integer, Vec[Integer]]].convert(idxs)
#             ).map(
#                 Abstraction[
#                     Pair[NDArray, Either[Integer, Vec[Integer]]], NDArray
#                 ].from_fn(lambda p: p.left[p.right])
#             )
#         )

#     @expression
#     def __add__(self, other: NDArrayCompat) -> NDArrayCompat:
#         ...

#     @expression
#     @classmethod
#     def from_ndarray(cls, n: Maybe[NDArray]) -> NDArrayCompat:
#         ...

#     @expression
#     def to_ndarray(self) -> numpy.ndarray:
#         ...


# def ndarray_getitem()

# @register_core
# @rule



# @register_unbox
# @rule
# def box_unbox_ndarray_compat(n: NDArray) -> R[numpy.ndarray]:
#     return (NDArrayCompat.from_ndarray(Maybe.just(n)).to_ndarray(), n.to_ndarray())


# @register_convert
# @rule
# def add_compat(l: NDArray, r: NDArray) -> R[NDArrayCompat]:
#     return (
#         NDArrayCompat.from_ndarray(Maybe.just(l))
#         + NDArrayCompat.from_ndarray(Maybe.just(r)),
#         NDArrayCompat.from_ndarray(Maybe.just(l + r)),
#     )


# register_convert(default_rule(NDArrayCompat.__getitem__))


# @register_convert
# @rule
# def convert_to_ndarray(i: Maybe[NDArray]) -> R[Maybe[NDArray]]:
#     return Converter[NDArray].convert(NDArrayCompat.from_ndarray(i)), i


# @expression
# def arange(stop: object) -> NDArrayCompat:
#     return NDArrayCompat.from_ndarray(
#         Converter[Integer].convert(stop).map(Abstraction.from_fn(arange_))
#     )


# register_convert(default_rule(arange))


# @expression
# def arange_(stop: Integer) -> NDArray:
#     ...



from __future__ import annotations
from metadsl import *
from metadsl_core import *
from metadsl_rewrite import *

from .int_compat import *
from .injest import *


zero = IntCompat.from_maybe_integer(Maybe.just(Integer.from_int(0)))
one = IntCompat.from_maybe_integer(Maybe.just(Integer.from_int(1)))


def test_injest() -> None:
    assert execute(injest(0)) == zero
    assert execute(injest(Integer.from_int(0))) == zero
    assert execute(injest(zero)) == zero


def test_add() -> None:
    assert execute(zero + 1) == one


def test_radd() -> None:
    assert execute(1 + zero) == one



from __future__ import annotations
import metadsl_all as m


class Number(m.Expression):
    @m.expression
    def __add__(self, other: Number) -> Number:
        ...

    @m.expression
    @classmethod
    def from_int(cls, i: int) -> Number:
        ...


@m.register_core
@m.rule
def add_zero(y: Number):
    yield Number.from_int(0) + y, y
    yield y + Number.from_int(0), y


assert m.execute(Number.from_int(0) + Number.from_int(10)) == Number.from_int(10)



"""
Rewrite rules for metadsl expressions
"""

from metadsl import export_from


from .normalize import *  # type: ignore
from .rules import *  # type: ignore
from .strategies import *  # type: ignore
from .combinators import *  # type: ignore


strategy = StrategyNormalize()
execute = Executor(strategy)
register = Registrator(strategy)

export_from(
    __name__,
    "normalize",
    "rules",
    "strategies",
    "combinators",
    local=["execute", "register"],
)

__version__ = "0.1.0"



"""
Pattern matching allows you to define replacement strategies by creating a template that should resemble
an expression, with certain leaf nodes as Wildcards. These can match any node in the target expression.

The values at these wildcards are captured and passed to another function to compute a resulting expression
to use as the replacement.

Conceptually, this is similar to regular expression matching / finite state transducers. Technically,
this is a form of symbolic tree transducers, but not implemented in a mathematically consistant way. If it was,
then we could more easily combine matches to make them execute faster and compute properties about them. This would
be good to add at a later date, and could be done without having users change their code.
"""
from __future__ import annotations

import dataclasses
import functools
import inspect
import logging
import types
import typing

import typing_inspect
from metadsl import *
from metadsl.typing_tools import *

from .strategies import *

__all__ = ["R", "NoMatch", "rule", "default_rule", "create_wildcard"]

T = typing.TypeVar("T")

R = typing.Union[
    typing.Tuple[T, typing.Union[typing.Callable[[], T], T]],
    typing.Generator[
        typing.Tuple[T, typing.Union[typing.Callable[[], T], T]], None, None
    ],
]
# Mapping from wildcards to the matching pattern and replacement
MatchFunctionType = typing.Callable[..., R]

logger = logging.getLogger(__name__)


class NoMatch(Exception):
    pass


def rule(fn: MatchFunctionType) -> Strategy:
    """
    Creates a new strategy given a callable that accepts wildcards and returns
    the match value and the replacement value.
    """

    return Rule(fn)


T_Callable = typing.TypeVar("T_Callable", bound=typing.Callable)


def default_rule(fn: typing.Callable) -> Strategy:
    """
    Creates a strategy based on the body of a passed in expression function
    """
    return DefaultRule(fn)


@dataclasses.dataclass(unsafe_hash=True)
class DefaultRule(Strategy):
    fn: typing.Callable
    inner_fn: typing.Callable = dataclasses.field(
        init=False, repr=False, compare=False, hash=False
    )

    def __str__(self):
        return f"{self.inner_fn.__module__}.{self.inner_fn.__qualname__}"

    def __post_init__(self):
        self.inner_fn = self.fn.__wrapped__  # type: ignore

    def optimize(self, executor, strategy):
        # TODO: Implement optimizations for default rules
        pass

    def __call__(self, ref: ExpressionReference) -> typing.Iterable[Result]:
        """
        This strategy should match whenever the expression is this function.

        Then, all it has to do is get the type variable mapping given the current args
        and apply it to the return value. This is so that if the body uses generic type
        variables, they are turned into the actual instantiations. 
        """
        with CaptureLogging() as results:
            expr = ref.expression
            logger.debug("DefaultRule.__call__ self=%s expr=%s", self, expr)
            if not isinstance(expr, Expression):
                return

            fn = self.fn

            args = expr.args

            # If any of the args are placeholders, don't match!
            if any(
                isinstance(arg, PlaceholderExpression)
                for arg in args + list(expr.kwargs.values())
            ):
                return None

            typevars: TypeVarMapping = infer_return_type(
                expr.function.fn,  # type: ignore
                getattr(expr.function, "owner", None),
                getattr(expr.function, "is_classmethod", False),
                tuple(args),
                expr.kwargs,
            )[-1]
            if isinstance(fn, BoundInfer) and isinstance(expr.function, BoundInfer):
                if fn.fn != expr.function.fn:
                    return None
                if fn.is_classmethod:
                    args = [
                        typing.cast(object, replace_typevars(typevars, fn.owner))
                    ] + args

            elif fn != expr.function:
                return None
            with TypeVarScope(*typevars.keys()):
                new_expr = self.inner_fn(*args, **expr.kwargs)
                result = ReplaceTypevarsExpression(typevars)(new_expr)
            logger.debug("DefaultRule.__call__ result=%s", result)

            ref.replace(result)
        yield Result(str(self), logs="\n".join(results))


class Wildcard(Expression, typing.Generic[T]):
    """
    Insert this somewhere in an expression tree to represent any expression or value of type T.
    """

    @expression
    @classmethod
    def create(cls, obj: object) -> T:
        """
        The object inside is only needed for identity checks, to different
        two wildcards if the objects are different.
        """
        ...


def create_wildcard(t: typing.Type[T]) -> T:
    """
    Returns a wildcard of type "t"
    """
    return Wildcard[t].create(object())  # type: ignore


# a number of (wildcard expression, replacement) pairs
WildcardMapping = typing.Mapping[Expression, object]


@dataclasses.dataclass(unsafe_hash=True)
class Rule(Strategy):
    """
    Creates a replacement strategy given a function that maps from wildcard inputs
    to two things, a template expression tree and a replacement thunk.

    If the template matches an expression, it will be replaced with the result of the thunk, replacing
    the input args with the nodes at their locations in the template.

    You can also return None from the strategy to signal that it won't match.
    """

    matchfunction: MatchFunctionType

    # the wildcards that are present in the template
    wildcards: typing.List[Expression] = dataclasses.field(
        init=False, hash=False, compare=False
    )

    results: typing.List[R] = dataclasses.field(init=False, hash=False, compare=False)

    def __str__(self):
        return f"{self.matchfunction.__module__}.{self.matchfunction.__qualname__}"

    def __post_init__(self):
        functools.update_wrapper(self, self.matchfunction)
        # Create one wildcard` per argument
        self.wildcards = [create_wildcard(a) for a in get_arg_hints(self.matchfunction)]

        # we match the wildcards against themselves to get an identity typevar mapping
        # TODO: Replace with just grabbing all typevars from arg types themselves
        typevars_in_args = match_values(
            tuple(self.wildcards), tuple(self.wildcards)
        ).keys()
        # Set the typevar args in scope so when functions are called here they are recorded properly
        with TypeVarScope(*typevars_in_args):
            # Call the function first to create a template with the wildcards
            result = self.matchfunction(*self.wildcards)
            self.results = (
                list(result)  # type: ignore
                if inspect.isgeneratorfunction(self.matchfunction)
                else [result]
            )

    def optimize(self, executor: Executor, strategy: Strategy) -> None:
        new_results: typing.List[R] = []

        for i, result in enumerate(self.results):
            template, expression_thunk = result
            # apply execution to all results that are not thunks
            if not isinstance(expression_thunk, types.FunctionType):
                self.results[i] = (template, executor(expression_thunk, strategy))

    def __call__(self, ref: ExpressionReference) -> typing.Iterable[Result]:
        expr = ref.expression
        with CaptureLogging() as logs:
            logger.debug("Rule.__call__ self=%s expr=%s", self, expr)
            for i, result in enumerate(self.results):
                template, expression_thunk = result
                try:
                    logger.debug("Trying to match against %s", template)
                    typevars, wildcards_to_nodes = match_expression(  # type: ignore
                        self.wildcards, template, expr
                    )
                except NoMatch:
                    logger.debug("Not a match")
                    continue
                logger.debug(
                    "Matched expr=%s typevars=%s", wildcards_to_nodes, typevars
                )
                # if the result is a function, we can't use substitution, so instead we re-call
                # with args and use that result
                if isinstance(expression_thunk, types.FunctionType):
                    args = [
                        wildcards_to_nodes.get(wildcard, wildcard)
                        for wildcard in self.wildcards
                    ]
                    with TypeVarScope(*typevars.keys()):
                        _, expression_thunk = (
                            list(self.matchfunction(*args))[i]
                            if inspect.isgeneratorfunction(self.matchfunction)
                            else self.matchfunction(*args)
                        )
                        # If it's a function, make sure we have real values instead of placeholders
                        # for any of the nodes that are placeholders for something more specific
                        # than a typevar, object, or any
                        if any(
                            isinstance(node, PlaceholderExpression)
                            and not is_vague_type(wildcard_inner_type(wildcard))
                            for wildcard, node in wildcards_to_nodes.items()
                        ):
                            continue
                        try:
                            result_expr: object = expression_thunk()
                        except NoMatch:
                            continue
                else:
                    result_expr = ReplaceValues(wildcards_to_nodes)(expression_thunk)
                with TypeVarScope(*typevars.keys()):
                    result_expr = ReplaceTypevarsExpression(typevars)(result_expr)
                logger.debug("Rule.__call__ res=%s", result_expr)
                ref.replace(result_expr)
                yield Result(
                    # if there is more than one possible match from this strategy, also put the index of the match
                    name=str(self) if len(self.results) == 1 else f"{self}[{i}]",
                    logs="\n".join(logs),
                )
                return


@dataclasses.dataclass
class ReplaceValues:
    mapping: typing.Mapping

    def __call__(self, expr):
        if expr in self.mapping:
            return self.mapping[expr]
        if not isinstance(expr, Expression):
            return expr
        result = expr._map(self)

        # if any of the args are create_iterated_placeholder
        # then remove those and replaced with arg expanded.
        new_args: typing.List = []
        for arg in result.args:
            if (
                isinstance(arg, Expression)
                and arg.function == create_iterated_placeholder
            ):
                inner_args = arg.args[0]
                assert isinstance(inner_args, tuple)
                for inner_arg in inner_args:
                    new_args.append(inner_arg)
            else:
                new_args.append(arg)
        result.args = new_args
        return result


@dataclasses.dataclass(frozen=True)
class ReplaceTypevarsExpression:
    """
    Make sure all functions have typevars attached to them, so when we call those functions
    the typevars will be replaced.

    Use a class instead of a function so we can partially apply it and have equality based on typevars 
    """

    typevars: TypeVarMapping

    def __call__(self, expression: object) -> object:
        """
        Replaces all typevars found in the classmethods of an expression.
        """
        typevars = self.typevars
        if isinstance(expression, Expression):
            return expression._map(
                fn=self,
                function_fn=lambda fn: replace_fn_typevars(fn, typevars),  # type: ignore
                type_fn=lambda tp: replace_typevars(typevars, tp),
            )
        return replace_fn_typevars(
            expression,
            typevars,
            ReplaceTypevarsExpression(typevars=HashableMapping(typevars)),
        )


def wildcard_inner_type(wildcard: object) -> typing.Type:
    """
    Return inner type for a wildcard
    """
    return typing_inspect.get_args(typing_inspect.get_generic_type(wildcard))[0]


def is_vague_type(t: typing.Type) -> bool:
    """
    Returns true if the type is a typevar, object or typing.
    """
    return typing_inspect.is_typevar(t) or t == object or t == typing.Any


def match_expression(
    wildcards: typing.List[Expression], template: object, expr: object
) -> typing.Tuple[TypeVarMapping, WildcardMapping]:
    """
    Returns a mapping of wildcards to the objects at that level, or None if it does not match.

    template: the expression with wildcards in it
    expr: the expression without wildcards to match against

    A wildcard can match either an expression or a value. If it matches two nodes, they must be equal.
    """
    logger.debug(
        "match_expression wildcards=%s template=%s expr=%s", wildcards, template, expr
    )
    if template in wildcards:
        logger.debug(
            "template is a wildcard, matching expr to template to get typevars"
        )
        # Match type of wildcard with type of expression
        try:
            res = (
                match_values(template, expr),
                UnhashableMapping(Item(typing.cast(Expression, template), expr)),
            )
            logger.debug("got wildcard mapping %s", res)
            return res
        except TypeError:
            logger.debug("could not match types")
            raise NoMatch

    if isinstance(expr, Expression):
        logger.debug("value is expression")
        if not isinstance(template, Expression):
            logger.debug("...but template isn't so no match")
            raise NoMatch
        # Any typevars in the template that are unbound should be matched with their
        # versions in the expr

        try:
            fn_type_mapping: TypeVarMapping = match_functions(
                template.function, expr.function
            )
        except TypeError:
            logger.debug("could not match functions")
            raise NoMatch
        logger.debug("matched functions to get typevar_apping=%s", fn_type_mapping)
        if set(expr.kwargs.keys()) != set(template.kwargs.keys()):
            logger.debug("No match because typevars not same keys")
            raise TypeError("Wrong kwargs in match")

        template_args: typing.Iterable[object]
        expr_args: typing.Iterable[object]

        # Process args in the template that can represent any number of args.
        # These are the "IteratedPlaceholder"s
        # Allow one iterated placeholder in the template args
        # For example fn(a, b, [...], *c, d, e, [...])
        # Here `c` should take as many args as it can between the ends,
        # Each of those should be matched against the inner
        iterated_args = [
            arg for arg in template.args if isinstance(arg, IteratedPlaceholder)
        ]
        if iterated_args:
            # Only support one iterated arg for now
            # TODO: Support more than one, would require branching
            assert len(iterated_args) == 1
            # template args, minus the iterators, is the minimum length of the values
            # If they have less values than this, raise an error
            min_n_args = len(template.args) - 1
            if len(expr.args) < min_n_args:
                raise NoMatch("Wrong number of args in match")
            (template_iterated,) = iterated_args
            template_index_iterated = template.args.index(template_iterated)

            template_args = list(template.args)
            # Swap template iterated with inner wildcard
            (template_args[template_index_iterated],) = template_iterated.args

            expr_args = collapse_tuple(
                tuple(expr.args),
                template_index_iterated,
                # The number we should preserve on the right, is the number of template
                # args after index
                len(template.args) - template_index_iterated - 1,
            )

        else:
            if len(template.args) != len(expr.args):
                raise NoMatch("Wrong number of args in match")
            template_args = template.args
            expr_args = expr.args

        type_mappings, expr_mappings = list(
            zip(
                *(
                    match_expression(wildcards, arg_template, arg_value)
                    for arg_template, arg_value in zip(template_args, expr_args)
                ),
                *(
                    match_expression(wildcards, template.kwargs[key], expr.kwargs[key])
                    for key in template.kwargs.keys()
                ),
            )
        ) or ((), ())
        logger.debug("Matched args and kwargs")
        try:
            merged_typevars: TypeVarMapping = merge_typevars(
                fn_type_mapping, *type_mappings
            )
        except TypeError:
            raise NoMatch
        try:
            return (
                merged_typevars,
                safe_merge(*expr_mappings, dict_constructor=UnhashableMapping),
            )
        except ValueError:
            raise NoMatch
    if template != expr:
        raise NoMatch
    return match_values(template, expr), UnhashableMapping()


def collapse_tuple(t: typing.Tuple, l: int, r: int) -> typing.Tuple:
    """
    Collapses the middle of a tuple into another nested tuple, keeping l on the left side
    uncollapsed and r on the right side un collapsed

    >>> collapse_tuple(tuple(range(10)), 3, 4)
    (0, 1, 2, (3, 4, 5), 6, 7, 8, 9)
    >>> collapse_tuple(tuple(range(5)), 3, 0)
    (0, 1, 2, (3, 4))
    """
    if r == 0:
        return tuple([*t[:l], t[l:]])
    return tuple([*t[:l], t[l:-r], *t[-r:]])



from __future__ import annotations
import typing
import typing_inspect

from metadsl import *
from metadsl_core import *
from metadsl.typing_tools import get_type
from metadsl_rewrite import *

from .bool_compat import *
from .injest import *
from .int_compat import *
from .boxing import *


__all__ = ["HomoTupleCompat"]


T = typing.TypeVar("T")
U = typing.TypeVar("U")


class HomoTupleCompat(Expression, typing.Generic[T, U]):
    """
    Should follow Python API for tuple with homogoneous types. 


    T is the outer type, when we are getting the item.
    U is the inner type of the items, that they will be stored as.

    """

    @expression
    @classmethod
    def from_maybe_vec(cls, v: Maybe[Vec[U]]) -> HomoTupleCompat[T, U]:
        ...

    @expression  # type: ignore
    @property
    def to_maybe_vec(self) -> Maybe[Vec[U]]:
        ...

    def __getitem__(self, idx: object) -> typing.Union[T, HomoTupleCompat[T, U]]:
        """
        Don't use in strongly typed code.
        """
        # If we infer it to an int, then use the getitem int
        # otherwise use the slice
        try:
            injested_idx = injest(idx)
        except NotImplementedError:
            pass
        else:
            if isinstance(injested_idx, IntCompat):
                return self.getitem_int(idx)
        return self.getitem_slice(idx)

    # TODO: Rename to single and selection
    @expression
    def getitem_int(self, idx: object) -> T:
        ...

    @expression
    def getitem_slice(self, idx: object) -> HomoTupleCompat[T, U]:
        ...

    def __setitem__(self, idx: object, value: object) -> None:
        res = self.setitem(idx, value)

        self.function = res.function  # type: ignore
        self.args = res.args
        self.kwargs = res.kwargs

    @expression
    def setitem(self, idx: object, value: object) -> HomoTupleCompat[T, U]:
        ...

    @expression  # type: ignore
    @property
    def length(self) -> IntCompat:
        ...

    @expression
    def __add__(self, other: object) -> HomoTupleCompat[T, U]:
        ...

    @expression
    def __radd__(self, other: object) -> HomoTupleCompat[T, U]:
        ...


@guess_type.register(HomoTupleCompat)
def guess_homo_tuple(ht: HomoTupleCompat[T, U]):
    compat_tp, inner_tp = typing_inspect.get_args(get_type(ht))
    return get_type(ht), Vec[inner_tp]  # type: ignore


@guess_type.register(Vec)
def guess_homo_tuple_vec(b: Vec[T]):
    (arg_tp,) = typing_inspect.get_args(get_type(b))
    compat_tp, inner_tp = guess_type_of_type(arg_tp)
    return HomoTupleCompat[compat_tp, inner_tp], Vec[inner_tp]  # type: ignore


@guess_type.register
def guess_homo_tuple_tuple(b: tuple):
    compat_tp, inner_tp = guess_types(*b)
    return HomoTupleCompat[compat_tp, inner_tp], Vec[inner_tp]  # type: ignore


@register_convert
@rule
def to_from_maybe_vec(v: Maybe[Vec[U]]) -> R[Maybe[Vec[U]]]:
    return HomoTupleCompat[T, U].from_maybe_vec(v).to_maybe_vec, v


@register_convert
@rule
def convert_to_vec(v: Maybe[Vec[U]]) -> R[Maybe[Vec[U]]]:
    return Converter[Vec[U]].convert(HomoTupleCompat[T, U].from_maybe_vec(v)), v


@register_convert
@rule
def box_homo_tuple(v: Maybe[Vec[U]]) -> R[HomoTupleCompat[T, U]]:
    return (
        Boxer[HomoTupleCompat[T, U], Vec[U]].box(v),
        HomoTupleCompat[T, U].from_maybe_vec(v),
    )


@register_convert
@rule
def setitem(v: Maybe[Vec[U]], idx: object, value: object) -> R[HomoTupleCompat[T, U]]:
    """
    Convert setitem to eiter integer and single item, or selection and vector
    """
    return (
        HomoTupleCompat[T, U].from_maybe_vec(v).setitem(idx, value),
        lambda: HomoTupleCompat[T, U].from_maybe_vec(
            (
                v
                & Converter[Either[Pair[Integer, U], Pair[Selection, Vec[U]]]].convert(
                    Pair.create(idx, value)
                )
            ).map(
                Abstraction[
                    Pair[Vec[U], Either[Pair[Integer, U], Pair[Selection, Vec[U]]]],
                    Vec[U],
                ].from_fn(
                    lambda xs: xs.right.match(
                        Abstraction[Pair[Integer, U], Vec[U]].from_fn(
                            lambda ys: xs.left.set(ys.left, ys.right)
                        ),
                        Abstraction[Pair[Selection, Vec[U]], Vec[U]].from_fn(
                            lambda ys: xs.left.set_selection(ys.left, ys.right)
                        ),
                    )
                )
            )
        ),
    )


@register_convert
@rule
def getitem_int(v: Maybe[Vec[U]], idx: object) -> R[T]:
    """
    Try to convert getitem to integer and index
    """
    return (
        HomoTupleCompat[T, U].from_maybe_vec(v).getitem_int(idx),
        lambda: Boxer[T, U].box(
            (v & Converter[Integer].convert(idx)).map(
                Abstraction[Pair[Vec[U], Integer], U].from_fn(
                    lambda xs: xs.left[xs.right]
                )
            )
        ),
    )


@register_convert
@rule
def getitem_slice(v: Maybe[Vec[U]], idx: object) -> R[HomoTupleCompat[T, U]]:
    return (
        HomoTupleCompat[T, U].from_maybe_vec(v).getitem_slice(idx),
        lambda: HomoTupleCompat[T, U].from_maybe_vec(
            (v & Converter[Selection].convert(idx)).map(
                Abstraction[Pair[Vec[U], Selection], Vec[U]].from_fn(
                    lambda xs: xs.left.select(xs.right)
                )
            )
        ),
    )


@register_convert
@rule
def length(v: Maybe[Vec[U]]) -> R[IntCompat]:
    return (
        HomoTupleCompat[T, U].from_maybe_vec(v).length,
        lambda: IntCompat.from_maybe_integer(
            v.map(Abstraction[Vec[U], Integer].from_fn(lambda v: v.length))
        ),
    )


@register_convert
@rule
def add(v: Maybe[Vec[U]], o: object) -> R[HomoTupleCompat[T, U]]:
    both_converted = v & Converter[Vec[U]].convert(o)
    # add
    yield (
        HomoTupleCompat[T, U].from_maybe_vec(v) + o,
        lambda: HomoTupleCompat[T, U].from_maybe_vec(
            both_converted.map(
                Abstraction[Pair[Vec[U], Vec[U]], Vec[U]].from_fn(
                    lambda both: both.left + both.right
                )
            )
        ),
    )
    # radd
    yield (
        o + HomoTupleCompat[T, U].from_maybe_vec(v),
        lambda: HomoTupleCompat[T, U].from_maybe_vec(
            both_converted.map(
                Abstraction[Pair[Vec[U], Vec[U]], Vec[U]].from_fn(
                    lambda both: both.right + both.left
                )
            )
        ),
    )



from __future__ import annotations

import dataclasses
import typing

from metadsl import *

from .strategies import *

__all__ = [
    "StrategyFold",
    "StrategySequence",
    "StrategyLabel",
    "StrategyRepeat",
    "StrategyInOrder",
]


@dataclasses.dataclass
class StrategyLabel(Strategy):
    """
    Takes in an existing strategy and makes the final state have a label.
    """

    label: str
    strategy: Strategy

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Result]:
        replaced = False
        for result in self.strategy(expr):
            replaced = True
            yield result
        if replaced:
            yield Result(name=self.label, label=self.label)

    def optimize(self, executor, strategy):
        self.strategy.optimize(executor, strategy)


@dataclasses.dataclass(init=False)
class StrategyInOrder(Strategy):
    """
    Returns a new replacement strategy that executes each of the strategies in order,
    returning a new expression if any of them replaced.
    """

    strategies: typing.Tuple[Strategy, ...]

    def __init__(self, *strategies: Strategy):
        self.strategies = strategies

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Result]:
        for strategy in self.strategies:
            for replacement in strategy(expr):
                yield replacement

    def optimize(self, executor, strategy):
        for strategy_ in self.strategies:
            strategy_.optimize(executor, strategy)


@dataclasses.dataclass(init=False)
class StrategySequence(Strategy):
    """
    Returns a new replacement strategy that tries each of the replacement strategies in sequence, returning the result of the first that matches.
    """

    strategies: typing.Tuple[Strategy, ...]

    def __init__(self, *strategies: Strategy):
        self.strategies = strategies

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Result]:
        for strategy in self.strategies:
            for replacement in strategy(expr):
                yield replacement
                return

    def optimize(self, executor, strategy):
        for strategy_ in self.strategies:
            strategy_.optimize(executor, strategy)


@dataclasses.dataclass
class StrategyFold(Strategy):
    """
    Returns the first replacement found by starting at the top of the expression tree
    and then recursing down into its leaves.
    """

    strategy: Strategy

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Result]:
        for child_ref in expr.descendents:
            for replacement in self.strategy(child_ref):
                yield replacement
                return

    def optimize(self, executor, strategy):
        self.strategy.optimize(executor, strategy)


@dataclasses.dataclass
class StrategyRepeat(Strategy):
    """
    Returns a new replacement strategy that repeatedly calls the replacement strategy
    """

    strategy: Strategy
    max_calls: int = 1000

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Result]:
        i = 0

        for i in range(self.max_calls):
            replaced = False
            for replacement in self.strategy(expr):
                replaced = True
                yield replacement
            if not replaced:
                return
        raise RuntimeError("Exceeded maximum number of repitions")

    def optimize(self, executor, strategy):
        """
        Apply passed in strategy repeatedly.
        """
        self.strategy.optimize(executor, self)



from __future__ import annotations
import typing

from metadsl import *
from metadsl_rewrite import *

from .abstraction import *
from .strategies import *
from .pair import *

__all__ = ["Maybe", "collapse_maybe"]
T = typing.TypeVar("T")
U = typing.TypeVar("U")
V = typing.TypeVar("V")


class Maybe(Expression, typing.Generic[T]):
    """
    An optional value.
    """

    @expression
    @classmethod
    def nothing(cls) -> Maybe[T]:
        ...

    @expression
    @classmethod
    def just(cls, value: T) -> Maybe[T]:
        ...

    @expression
    def match(self, nothing: U, just: Abstraction[T, U]) -> U:
        ...

    @expression
    def default(self, value: T) -> T:
        return self.match(value, Abstraction[T, T].from_fn(lambda i: i))

    @expression
    def __or__(self, other: Maybe[T]) -> Maybe[T]:
        """
        Like the <|> function https://en.wikibooks.org/wiki/Haskell/Alternative_and_MonadPlus
        """
        ...

    @expression
    def __and__(self, other: Maybe[U]) -> Maybe[Pair[T, U]]:
        """
        >>> execute(Maybe[int].nothing() & Maybe.just("")) == Maybe[Pair[int, str]].nothing()
        True
        >>> execute(Maybe.just(10) & Maybe[str].nothing()) == Maybe[Pair[int, str]].nothing()
        True
        >>> execute(Maybe.just(10) & Maybe.just("")) == Maybe.just(Pair.create(10, ""))
        True
        """
        ...

    @expression
    def map(self, just: Abstraction[T, U]) -> Maybe[U]:
        return self.match(
            Maybe[U].nothing(),
            Abstraction[U, Maybe[U]].from_fn(lambda v: Maybe.just(v)) + just,  # type: ignore
        )

    @expression
    def flat_map(self, just: Abstraction[T, Maybe[U]]) -> Maybe[U]:
        return collapse_maybe(self.map(just))


@expression
def collapse_maybe(x: Maybe[Maybe[T]]) -> Maybe[T]:
    return x.match(
        Maybe[T].nothing(), Abstraction[Maybe[T], Maybe[T]].from_fn(lambda m: m)
    )


register_core(default_rule(Maybe[T].map))
register_core(default_rule(Maybe[T].default))
register_core(default_rule(Maybe[T].flat_map))
register_core(default_rule(collapse_maybe))


@register_core  # type: ignore
@rule
def maybe_match(nothing: U, just: Abstraction[T, U], v: T) -> R[U]:
    yield Maybe[T].nothing().match(nothing, just), nothing
    yield Maybe.just(v).match(nothing, just), just(v)


@register_core
@rule
def maybe_or(x: T, v: Maybe[T]) -> R[Maybe[T]]:
    yield Maybe[T].nothing() | Maybe[T].nothing(), Maybe[T].nothing()
    yield Maybe.just(x) | v, Maybe.just(x)
    yield v | Maybe.just(x), Maybe.just(x)


@register_core
@rule
def maybe_and(left: Maybe[T], right: Maybe[U], left_v: T, right_v: U) -> R[Pair[T, U]]:
    yield Maybe[T].nothing() & right, Maybe[Pair[T, U]].nothing()
    yield left & Maybe[U].nothing(), Maybe[Pair[T, U]].nothing()
    yield Maybe.just(left_v) & Maybe.just(right_v), Maybe.just(
        Pair.create(left_v, right_v)
    )


@register_core
@rule
def push_down_maybe_maybe_match(
    m: Maybe[T],
    nothing: Maybe[U],
    just: Abstraction[T, Maybe[U]],
    nothing_outer: V,
    just_outer: Abstraction[U, V],
) -> R[V]:
    """
    If we have two matches, one after the other, push down the outer match inside the inner one.
    """
    return (
        m.match(nothing, just).match(nothing_outer, just_outer),
        lambda: m.match(
            nothing.match(nothing_outer, just_outer),
            Abstraction[T, V].from_fn(
                lambda t: just(t).match(nothing_outer, just_outer)
            ),
        ),
    )


@register_core
@rule
def match_identity(m: Maybe[T], var: T,) -> R[Maybe[T]]:
    """
    If we have a match that has each branch just create the same maybe, then remove the match.
    """
    return (m.match(Maybe[T].nothing(), Abstraction.create(var, Maybe.just(var))), m)



"""
Pattern matching strategies.

Inspired by this poster
    https://drive.google.com/file/d/10nWoSHGH1ACgdslcnPXwPvEekh9ae9x4/view
and this paper
    Visser et. al.: Building program optimizers with rewriting strategies (ICFP 1998)
"""
from __future__ import annotations

import typing
import dataclasses
from metadsl import *

__all__ = ["Strategy", "Executor", "Result"]

T = typing.TypeVar("T")


def _execute_all(ref: ExpressionReference, strategy: Strategy) -> object:
    """
    Call a replacement strategy many times, returning the last result whole.
    """
    for replacement in strategy(ref):
        pass
    return ref.expression


@dataclasses.dataclass
class Executor:
    # rule that is used if none is passed in
    default_strategy: Strategy
    # Function that takes a rule and an expression and executes it
    execute: typing.Callable[
        [ExpressionReference, Strategy], object
    ] = dataclasses.field(default=_execute_all)

    def __call__(self, expr: T, strategy: typing.Optional[Strategy] = None) -> T:
        execute: typing.Callable[  # type: ignore
            [ExpressionReference, Strategy], object
        ] = self.execute  # type: ignore
        default_strategy: Strategy = self.default_strategy  # type: ignore
        strategy = strategy or default_strategy
        assert strategy
        return typing.cast(
            T,
            execute(
                ExpressionReference.from_expression(clone_expression(expr)), strategy
            ),
        )

    def optimize(self) -> None:
        self.default_strategy.optimize(self, NoOpStrategy())


@dataclasses.dataclass
class Result:
    # The name of the strategy that was executed
    name: str
    # Any optional logs for this strategy
    logs: str = ""
    # An optional label for this replacement
    label: typing.Optional[str] = None


class Strategy(typing.Protocol):
    def __call__(self, ref: ExpressionReference) -> typing.Iterable[Result]:
        ...

    def optimize(self, executor: Executor, strategy: Strategy) -> None:
        """

        """
        ...


class NoOpStrategy(Strategy):
    def __call__(self, ref):
        return tuple()

    def optimize(self, executor, strategy):
        pass



from metadsl_rewrite import *

__all__ = ["register_core", "register_ds", "register_convert"]

register_core = register["core"]
register_convert = register["conversion"]
register_ds = register["data structures"]



# import typing

# import numpy
# from metadsl import *

# from . import numpy as numpy_api
# from .abstraction import *
# from .conversion import *
# from .either import *
# from .integer import *
# from .maybe import *
# from .pair import *
# from .strategies import *
# from .vec import *

# __all__ = ["ndarray_getitem"]

# T = typing.TypeVar("T")


# @expression
# def unbox_integer(i: Integer) -> int:
#     ...


# @register_unbox  # type: ignore
# @rule
# def unbox_integer_rule(i: int) -> R[int]:
#     return unbox_integer(Integer.from_int(i)), i


# @expression
# def arange(stop: int) -> numpy.ndarray:
#     return numpy.arange(stop)


# register_numpy_engine(default_rule(arange))


# @register_unbox
# @rule
# def unbox_arange(i: Integer) -> R[numpy.ndarray]:
#     return (numpy_api.arange_(i).to_ndarray(), arange(unbox_integer(i)))


# @expression
# def ndarray_getitem(
#     self: numpy.ndarray, i: typing.Union[int, typing.Tuple[int, ...]]
# ) -> numpy.ndarray:
#     # Always return array not scalars
#     return numpy.array(self[i], ndmin=0)


# register_numpy_engine(default_rule(ndarray_getitem))


# @expression
# def ndarray_add(self: numpy.ndarray, other: numpy.ndarray) -> numpy.ndarray:
#     return self + other


# register_numpy_engine(default_rule(ndarray_add))


# @register_unbox  # type: ignore
# @rule
# def unbox_ndarray_add(a: numpy_api.NDArray, b: numpy_api.NDArray) -> R[numpy.ndarray]:
#     return (a + b).to_ndarray(), ndarray_add(a.to_ndarray(), b.to_ndarray())


# @expression
# def unbox_idxs(
#     idxs: Either[Integer, Vec[Integer]]
# ) -> typing.Union[int, typing.Tuple[int, ...]]:
#     ...


# @expression
# def homo_tuple(*values: T) -> typing.Tuple[T, ...]:
#     return tuple(values)


# register_core(default_rule(homo_tuple))


# @register_unbox  # type: ignore
# @rule
# def unbox_idxs_rule(
#     i: Integer, ints: typing.Sequence[Integer]
# ) -> R[typing.Union[int, typing.Tuple[int, ...]]]:
#     yield (unbox_idxs(Either[Integer, Vec[Integer]].left(i)), unbox_integer(i))
#     yield (
#         unbox_idxs(Either[Integer, Vec[Integer]].right(Vec.create(*ints))),
#         lambda: homo_tuple(*(unbox_integer(i_) for i_ in ints)),  # type: ignore
#     )


# @register_unbox  # type: ignore
# @rule
# def unbox_ndarray_getitem(
#     a: numpy_api.NDArray, idx: Either[Integer, Vec[Integer]]
# ) -> R[numpy.ndarray]:
#     return a[idx].to_ndarray(), ndarray_getitem(a.to_ndarray(), unbox_idxs(idx))




"""
LLVM support.
"""

__version__ = "0.3.0"


from metadsl import export_from

from .ctypes import *
from .integration import *
from .ir import *
from .ir_context import *
from .ir_llvmlite import *
from .llvmlite_binding import *
from .strategies import *

export_from(
    __name__,
    "ctypes",
    "integration",
    "ir_llvmlite",
    "ir",
    "llvmlite_binding",
    "ir_context",
    "strategies",
)



# """
# Multidimensional indices

# from https://gist.github.com/asmeurer/4913d32da667690e49f4675822219cb3
# """
# from __future__ import annotations

# from metadsl import *
# import typing
# import numpy





# __all__ = ["NDIndex", "NumPyIndex"]

# NumPyIndexInnerType = typing.Union[slice, ellipsis, int]
# NumPyIndexType = typing.Union[NumPyIndexInnerType, typing.Tuple[NumPyIndexInnerType, ...]]


# PossibleWrappedType = 'typing.Union[NumPyIndexInnerType, NDIndexCompat]'

# class NDIndexCompat(Expression):
#     """
#     Multi dimensional index that when used on some array returns some subset
#     of the values
#     """

#     @expression
#     @classmethod
#     def wrap(cls, v: NumPyIndexType) -> NDIndex:
#         ...


#     @expression
#     def __getitem__(self, v: PossibleWrappedType) -> NDIndexCompat:
#         """

#         """
#         ...


#     @expression
#     def __add__(self, v: PossibleWrappedType) -> NDIndexCompat:
#         ...

#     @expression
#     def len(self)

#     @expression
#     def length


from __future__ import annotations
import logging
import dataclasses
import typing

__all__ = ["toggle_debug_logging", "CaptureLogging"]


def metadsl_filter(record: logging.LogRecord) -> bool:
    """
    Only allow through if in metadsl
    """
    return record.name.startswith("metadsl")


handler = logging.StreamHandler()
handler.addFilter(metadsl_filter)


def toggle_debug_logging(enabled: bool) -> None:
    logging.basicConfig(  # type: ignore
        level=logging.DEBUG if enabled else logging.WARNING,
        force=True,
        handlers=[handler],
    )


class CapturingHandler(logging.Handler):
    """
    A logging handler capturing all (raw and formatted) logging output.
    """

    def __init__(self, events: typing.List[str]):
        super().__init__()
        self.events = events

    def emit(self, record):
        self.events.append(logging.Formatter("%(name)s: %(message)s").format(record))


@dataclasses.dataclass
class CaptureLogging:
    """
    Capture all logging.

        with CaputureLogging() as results:
            ...
        # results contains a list of all log messages, formatted
    

    Some inspiration from unittest.assertLogs
    """

    events: typing.List[str] = dataclasses.field(default_factory=list)
    old_handlers: typing.List[logging.Handler] = dataclasses.field(init=False)

    def __enter__(self):
        self.old_handlers = logging.root.handlers
        logging.root.handlers = [CapturingHandler(self.events)]
        return self.events

    def __exit__(self, et, ev, tb):
        logging.root.handlers = self.old_handlers



from __future__ import annotations

import typing
import dataclasses
import collections
import functools
import contextlib

from metadsl import *
from .strategies import *
from .combinators import *

__all__ = ["StrategyNormalize", "Registrator"]

T = typing.TypeVar("T")


def add_set(set_: typing.Set[T], item: T) -> T:
    set_.add(item)
    return item


@dataclasses.dataclass
class Registrator:
    """
    Helper class for adding strategies to the StrategyNormalize
    """

    strategy: StrategyNormalize

    def pre(self, strategy: Strategy) -> Strategy:
        self.strategy.pre.add(strategy)
        return strategy

    def post(self, strategy: Strategy) -> Strategy:
        self.strategy.post.add(strategy)
        return strategy

    def __getitem__(self, label: str) -> typing.Callable[[Strategy], Strategy]:
        set_ = self.strategy.phases[label]
        return functools.partial(add_set, set_)  # type: ignore

    def __delitem__(self, label: str) -> None:
        del self.strategy.phases[label]

    def __contains__(self, label: str) -> bool:
        return label in self.strategy.phases

    @contextlib.contextmanager
    def tmp(self, *strategies: Strategy):
        """
        Context manager that temporarily appends a list of strategies
        """
        TMP_NAME = "__tmp"
        for s in strategies:
            self[TMP_NAME](s)
        try:
            yield
        except Exception:
            raise
        finally:
            if TMP_NAME in self:
                del self[TMP_NAME]


@dataclasses.dataclass
class StrategyNormalize(Strategy):
    """
    A strategy combinator that expresses strategies in a number of phases. Each phase can
    execute strategies in all the phases before it. After each phase has no more matches,
    a label is set. This lets you explore when each phase is done executing.
    
    """

    pre: typing.Set[Strategy] = dataclasses.field(default_factory=set)
    post: typing.Set[Strategy] = dataclasses.field(default_factory=set)
    phases: typing.DefaultDict[str, typing.Set[Strategy]] = dataclasses.field(
        default_factory=lambda: collections.defaultdict(set)
    )

    def __call__(self, expr: ExpressionReference) -> typing.Iterable[Result]:
        return self.strategy(expr)

    def optimize(self, executor: Executor, strategy: Strategy) -> None:
        self.strategy.optimize(executor, strategy)
        # """
        # pre and post stay the same and the strategies in phases are all updated
        # """
        # assert not strategy
        # new_phases = collections.OrderedDict[str, typing.Set[Strategy]]()
        # for old_phase, strategy in zip(self.phases.items(), self.phase_strategies):
        #     label, old_strategies = old_phase
        #     new_phases[label] = set(
        #         old_strategy.optimize(executor, strategy)
        #         for old_strategy in old_strategies
        #     )
        # return dataclasses.replace(self, phases=new_phases)

    @property
    def phase_strategies(self) -> typing.Iterable[Strategy]:
        current_strategies: typing.Set[Strategy] = set()
        for label, strategies in self.phases.items():
            current_strategies.update(strategies)
            yield StrategyLabel(
                label,
                StrategyRepeat(
                    StrategySequence(
                        StrategyFold(StrategySequence(*self.pre)),
                        StrategyFold(StrategySequence(*current_strategies)),
                    )
                ),
            )

    @property
    def strategy(self) -> Strategy:
        return StrategyRepeat(
            StrategyInOrder(
                StrategyLabel(
                    "pre", StrategyRepeat(StrategyFold(StrategySequence(*self.pre)))
                ),
                *self.phase_strategies,
                StrategyLabel("post", StrategyFold(StrategySequence(*self.post)))
            )
        )



"""
Array tools built on metadsl

On the user end, we have a Array.

We should be able to do getitem, addition, concat (many arrays with same shape),
and chunk (split it into many arrays with same shape).

It will have a create(shape, idx_fn), shape


We will use a `assertInRange(int, low, hi)` to create bounded integers and do math on those.


We add an ArrayManager class that has a getitem that takes an integer and returns
a np array and has a `persist` method that takes an
array and gives a new one. It will be created with a list of NP arrays
as well as a dict that allows unhashable keys (using equality for testing)
that maps hashes of expression to indexes.  It

It will:

1. splits into chunks of length N
2. Map these arrays with the persist_single function which
   takes the called index fn (with some placeholder) and shape and returns a new array.
3. In a new phase of the replacement, so that the index fn has time to be simplified,
   look it up in the dict. If it's there, replace with the array with `np_array(manager[i])`
   If it's not there, replace with `np_array(manager.save_chunk(arr))`
3. In a new phase replace `manager.save_chunk(arr)` by creating a new empty array, and save it as `manager.saving_chunk(i, *indices)`
   where indices are a indices are a `np_setitem(arr, to_int_tuple(idxs), to_int(value))` which rerturn non


`np_array(x)` translate to `np_getitem(x, to_int_tuple(idxs))`

`manager.saving_chink(i, None, *)` -> `manager.saving_chunk(i, *)`
`manager.saving_chunk(i)` -> `manager[i]`


NO WAIT. I can't generate all these... I must generate ast :(




"""

__version__ = "0.0.0"

from metadsl import export_from

from .bool_compat import *
from .boxing import *
from .function_compat import *
from .injest import *
from .int_compat import *
from .tuple_compat import *

export_from(
    __name__,
    "injest",
    "boxing",
    "bool_compat",
    "int_compat",
    "tuple_compat",
    "function_compat",
)



from metadsl_rewrite import *

__all__ = [
    "register_context",
    "register_integration",
    "register_llvmlite",
    "register_ctypes",
    "register_llvmlite_binding",
]

# First do higher level context rules
register_context = register["llvmlite.context"]
# First do integration rules so that all functions are created
register_integration = register["llvmlite.integration"]
# Then create the llvm modules
register_llvmlite = register["llvmlite.ir_llvmlite"]
#  Finally you can create the ctypes
register_ctypes = register["llvmlite.ctypes"]
# And then compile into a function
register_llvmlite_binding = register["llvmlite.llvmlite_binding"]



from __future__ import annotations

import typing

from metadsl import *

from metadsl_rewrite import *

from .abstraction import *
from .conversion import *
from .maybe import *
from .strategies import *

__all__ = ["Boolean"]

T = typing.TypeVar("T")
U = typing.TypeVar("U")


class Boolean(Expression):
    # TODO: Remove this
    @expression
    @classmethod
    def create(cls, b: bool) -> Boolean:
        ...

    @expression
    def if_(self, true: T, false: T) -> T:
        ...

    @expression
    def and_(self, other: Boolean) -> Boolean:
        ...

    @expression
    def or_(self, other: Boolean) -> Boolean:
        ...

    @expression
    @classmethod
    def true(cls) -> Boolean:
        return Boolean.create(True)

    @expression
    @classmethod
    def false(cls) -> Boolean:
        return Boolean.create(False)


register_ds(default_rule(Boolean.true))
register_ds(default_rule(Boolean.false))


@register_ds  # type: ignore
@rule
def if_(b: bool, l: T, r: T) -> R[T]:
    return Boolean.create(b).if_(l, r), lambda: l if b else r


@register_ds  # type: ignore
@rule
def and_or(l: bool, r: bool) -> R[Boolean]:
    l_boolean = Boolean.create(l)
    r_boolean = Boolean.create(r)

    yield l_boolean.and_(r_boolean), lambda: Boolean.create(l and r)
    yield l_boolean.or_(r_boolean), lambda: Boolean.create(l or r)


@register_convert
@rule
def convert_bool(b: bool) -> R[Maybe[Boolean]]:
    """
    >>> execute(Converter[Boolean].convert(True), convert_bool) == Maybe.just(Boolean.create(True))
    True
    >>> list(convert_bool(ExpressionReference.from_expression(Converter[Boolean].convert("not bool"))))
    []
    """
    return Converter[Boolean].convert(b), Maybe.just(Boolean.create(b))


@register_core
@rule
def pull_if_above_maybe(
    b: Boolean,
    maybe_true: Maybe[T],
    maybe_false: Maybe[T],
    nothing: U,
    just: Abstraction[T, U],
):
    """
    Pull up if opreator around match... maybe could also pull this up generically on other operators?

    Do we want to pull if up always?
    """

    return (
        b.if_(maybe_true, maybe_false).match(nothing, just),
        b.if_(maybe_true.match(nothing, just), maybe_false.match(nothing, just)),
    )



"""
Core data types and replacements for metadsl
"""

from metadsl import export_from

from .abstraction import *
from .boolean import *
from .conversion import *
from .either import *
from .function import *
from .integer import *
from .maybe import *
from .pair import *
from .strategies import *
from .vec import *

export_from(
    __name__,
    "abstraction",
    "boolean",
    "conversion",
    "either",
    "function",
    "integer",
    "maybe",
    "pair",
    "strategies",
    "vec",
)

__version__ = "0.4.0"



from __future__ import annotations
import typing


from metadsl import *
from metadsl_core import *
from metadsl_rewrite import *
from .injest import *
from .boxing import *

__all__ = ["BoolCompat"]

T = typing.TypeVar("T")
U = typing.TypeVar("U")


class BoolCompat(Expression):
    """
    API should be work like Python's `bool` type.
    """

    @expression
    @classmethod
    def from_maybe_boolean(cls, i: Maybe[Boolean]) -> BoolCompat:
        ...

    @expression
    def and_(self, other: object) -> BoolCompat:
        ...

    @expression
    def or_(self, other: object) -> BoolCompat:
        ...

    def if_(self, l: object, r: object) -> object:
        compat_tp, inner_tp = guess_types(l, r)
        boxer = Boxer[compat_tp, inner_tp]  # type: ignore
        return boxer.box(self.if_maybe(boxer.convert(l), boxer.convert(r)))

    @expression
    def if_maybe(self, l: Maybe[T], r: Maybe[T]) -> Maybe[T]:
        ...


@guess_type.register(BoolCompat)
@guess_type.register(Boolean)
@guess_type.register(bool)
def guess_bool_compat(b: object):
    return BoolCompat, Boolean


@register_convert
@rule
def box_boolean(b: Maybe[Boolean]) -> R[BoolCompat]:
    return Boxer[BoolCompat, Boolean].box(b), BoolCompat.from_maybe_boolean(b)


@register_convert
@rule
def convert_bool_compat(b: Maybe[Boolean]) -> R[Maybe[Boolean]]:
    return Converter[Boolean].convert(BoolCompat.from_maybe_boolean(b)), b


@register_convert
@rule
def and_or(maybe_l: Maybe[Boolean], r: object) -> R[BoolCompat]:
    maybe_r = Converter[Boolean].convert(r)
    maybe = maybe_l & maybe_r
    compat_l = BoolCompat.from_maybe_boolean(maybe_l)

    def create_result(m: typing.Callable[[Boolean, Boolean], Boolean]) -> BoolCompat:
        return BoolCompat.from_maybe_boolean(
            maybe.map(
                Abstraction[Pair[Boolean, Boolean], Boolean].from_fn(
                    lambda p: m(p.left, p.right)
                )
            )
        )

    yield (
        compat_l.and_(r),
        create_result(lambda l_bool, r_bool: l_bool.and_(r_bool)),
    )

    yield (
        compat_l.or_(r),
        create_result(lambda l_bool, r_bool: l_bool.or_(r_bool)),
    )


@register_convert  # type: ignore
@rule
def if_maybes(cond: Maybe[Boolean], true: Maybe[T], false: Maybe[T]) -> R[Maybe[T]]:
    return (
        BoolCompat.from_maybe_boolean(cond).if_maybe(true, false),
        (cond & (true & false)).map(
            Abstraction[Pair[Boolean, Pair[T, T]], T].from_fn(
                lambda p: p.left.if_(*p.right.spread)
            )
        ),
    )



import pytest
from metadsl import *
from metadsl_rewrite import *

from .abstraction import *
from .conversion import *
from .maybe import *
from .strategies import *


class TestConvertIdentity:
    def test_matches_type(self):
        assert execute(Converter[int].convert(1)) == Maybe.just(1)

    def test_doesnt_match_type(self):
        assert not list(
            convert_identity_rule(
                ExpressionReference.from_expression(Converter[str].convert(1))
            )
        )

    def test_matches_convert(self):
        assert execute(Converter[int].convert(Maybe.just(1))) == Maybe.just(1)
        assert (
            execute(Converter[int].convert(Maybe[int].nothing()))
            == Maybe[int].nothing()
        )


class TestConvertToMaybe:
    def test_just(self):
        assert execute(Converter[Maybe[int]].convert(1)) == Maybe.just(Maybe.just(1))

    def test_nothing(self):
        assert execute(Converter[Maybe[int]].convert(None)) == Maybe.just(
            Maybe[int].nothing()
        )


class T(Expression):
    ...


class U(Expression):
    ...


class V(Expression):
    ...


class X(Expression):
    ...


@expression
def t_to_u(t: T) -> U:
    ...


@expression
def v() -> V:
    ...


@expression
def v_to_t(v: V) -> T:
    ...


@expression
def u_to_x(u: U) -> X:
    ...


@rule
def convert_v_to_t_just(v: V) -> R[Maybe[T]]:
    return Converter[T].convert(v), Maybe.just(v_to_t(v))


@rule
def convert_v_to_t_nothing(v: V) -> R[Maybe[T]]:
    return Converter[T].convert(v), Maybe[T].nothing()


@rule
def convert_u_to_x_just(u: U) -> R[Maybe[X]]:
    return Converter[X].convert(u), Maybe.just(u_to_x(u))


@rule
def convert_u_to_x_nothing(u: U) -> R[Maybe[X]]:
    return Converter[X].convert(u), Maybe[X].nothing()


class TestConvertToAbstraction:
    @pytest.mark.parametrize(
        "u_to_x_rule",
        [convert_v_to_t_just, convert_u_to_x_nothing],
        ids=["just", "nothing"],
    )
    @pytest.mark.parametrize(
        "v_to_t_rule",
        [convert_v_to_t_just, convert_v_to_t_nothing],
        ids=["just", "nothing"],
    )
    def test_from_abstraction(self, u_to_x_rule, v_to_t_rule) -> None:
        """
        Test converting from a function to an abstraction
        """

        def fn(t: T) -> U:
            return t_to_u(t)

        converted_a = Converter[Abstraction[V, Maybe[X]]].convert(fn)
        called_with_v: Maybe[X] = converted_a.flat_map(
            Abstraction[Abstraction[V, Maybe[X]], Maybe[X]].from_fn(lambda a: a(v()))
        )

        desired_result: Maybe[X] = (
            Converter[T]
            .convert(v())
            .flat_map(
                Abstraction[T, Maybe[X]].from_fn(
                    lambda t: Converter[X].convert(t_to_u(t))
                )
            )
        )
        with register.tmp(v_to_t_rule, u_to_x_rule):
            assert execute(called_with_v) == execute(desired_result)



from __future__ import annotations
from metadsl import *
from metadsl_core import *
from metadsl_rewrite import *

from .int_compat import *
from .tuple_compat import *
from .injest import *


item = IntCompat.from_maybe_integer(Maybe.just(Integer.from_int(0)))
tpl = HomoTupleCompat[IntCompat, Integer].from_maybe_vec(
    Maybe.just(Vec.create(Integer.from_int(0)))
)


def test_injest() -> None:
    assert execute(injest((0,))) == tpl
    # assert execute(injest(Integer.from_int(0))) == zero
    assert execute(injest(tpl)) == tpl


def test_getitem_int() -> None:
    assert execute(tpl[0]) == item




import importlib
import typing
import sys
__all__ = ["export_from"]


def export_from(module_name: str, *submodules: str, local: typing.List[str]=[]) -> None:
    """
    Sets the  `__all__` for the module to be the union of all the __all__'s of the submodules,
    plus the locals.

    Needed so that the modules themselves are not present in the __init__.py file, but all the exports are.
    """
    m = sys.modules[module_name]
    m.__all__ = list(local) #type: ignore
    for submodule in submodules:
        m.__all__.extend(importlib.import_module("." + submodule, module_name).__all__)#type: ignore


from __future__ import annotations
import typing


from metadsl import *
from metadsl_core import *
from .bool_compat import *
from .injest import *
from .boxing import *
from metadsl_rewrite import *

__all__ = ["IntCompat"]


class IntCompat(Expression):
    """
    API should be work like Python's `int` type.
    """

    @expression
    @classmethod
    def from_maybe_integer(cls, i: Maybe[Integer]) -> IntCompat:
        ...

    # Numeric
    @expression
    def __add__(self, other: object) -> IntCompat:
        ...

    @expression
    def __sub__(self, other: object) -> IntCompat:
        ...

    @expression
    def __mul__(self, other: object) -> IntCompat:
        ...

    @expression
    def __floordiv__(self, other: object) -> IntCompat:
        ...

    @expression
    def __mod__(self, other: object) -> IntCompat:
        ...

    # Right numeric

    @expression
    def __radd__(self, other: object) -> IntCompat:
        ...

    @expression
    def __rsub__(self, other: object) -> IntCompat:
        ...

    @expression
    def __rmul__(self, other: object) -> IntCompat:
        ...

    @expression
    def __rfloordiv__(self, other: object) -> IntCompat:
        ...

    @expression
    def __rmod__(self, other: object) -> IntCompat:
        ...

    # Comparison
    @expression
    def eq(self, other: object) -> BoolCompat:
        ...

    @expression
    def __lt__(self, other: object) -> BoolCompat:
        ...

    @expression
    def __le__(self, other: object) -> BoolCompat:
        ...

    @expression
    def __gt__(self, other: object) -> BoolCompat:
        ...

    @expression
    def __ge__(self, other: object) -> BoolCompat:
        ...


@guess_type.register(IntCompat)
@guess_type.register(Integer)
@guess_type.register(int)
def guess_int(i: object):
    return IntCompat, Integer


@register_convert
@rule
def convert_to_integer(i: Maybe[Integer]) -> R[Maybe[Integer]]:
    return Converter[Integer].convert(IntCompat.from_maybe_integer(i)), i


@register_convert
@rule
def box_int_compat(x: Maybe[Integer]) -> R[IntCompat]:
    return (
        Boxer[IntCompat, Integer].box(x),
        IntCompat.from_maybe_integer(x),
    )


@register_convert
@rule
def int_compat_operators(maybe_l: Maybe[Integer], r: object) -> R[IntCompat]:
    maybe_r = Converter[Integer].convert(r)
    maybe = maybe_l & maybe_r
    compat_l = IntCompat.from_maybe_integer(maybe_l)

    def create_int_result(
        m: typing.Callable[[Integer, Integer], Integer]
    ) -> typing.Callable[[], IntCompat]:
        return lambda: IntCompat.from_maybe_integer(
            maybe.map(
                Abstraction[Pair[Integer, Integer], Integer].from_fn(
                    lambda p: m(p.left, p.right)
                )
            )
        )

    def create_bool_result(
        m: typing.Callable[[Integer, Integer], Boolean]
    ) -> typing.Callable[[], BoolCompat]:
        return lambda: BoolCompat.from_maybe_boolean(
            maybe.map(
                Abstraction[Pair[Integer, Integer], Boolean].from_fn(
                    lambda p: m(p.left, p.right)
                )
            )
        )

    # numeric
    yield (
        compat_l + r,
        create_int_result(lambda l_int, r_int: l_int + r_int),
    )
    yield (
        compat_l - r,
        create_int_result(lambda l_int, r_int: l_int - r_int),
    )
    yield (
        compat_l * r,
        create_int_result(lambda l_int, r_int: l_int * r_int),
    )
    yield (
        compat_l // r,
        create_int_result(lambda l_int, r_int: l_int // r_int),
    )
    yield (
        compat_l % r,
        create_int_result(lambda l_int, r_int: l_int % r_int),
    )

    # right numeric
    yield (
        r + compat_l,
        create_int_result(lambda l_int, r_int: r_int + l_int),
    )
    yield (
        r - compat_l,
        create_int_result(lambda l_int, r_int: r_int - l_int),
    )
    yield (
        r * compat_l,
        create_int_result(lambda l_int, r_int: r_int * l_int),
    )
    yield (
        r // compat_l,
        create_int_result(lambda l_int, r_int: r_int // l_int),
    )
    yield (
        r % compat_l,
        create_int_result(lambda l_int, r_int: r_int % l_int),
    )

    # comparison
    yield (
        compat_l.eq(r),
        create_bool_result(lambda l_int, r_int: l_int.eq(r_int)),
    )
    yield (
        compat_l < r,
        create_bool_result(lambda l_int, r_int: l_int < r_int),
    )
    yield (
        compat_l <= r,
        create_bool_result(lambda l_int, r_int: l_int <= r_int),
    )
    yield (
        compat_l > r,
        create_bool_result(lambda l_int, r_int: l_int > r_int),
    )
    yield (
        compat_l >= r,
        create_bool_result(lambda l_int, r_int: l_int >= r_int),
    )
